#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Makolator Testing."""

from pathlib import Path

from makolator import Config, Existing

COMMENT_MAP = {
    ".c": "//",
    ".c++": "//",
    ".cpp": "//",
    ".ini": "#",
    ".py": "#",
    ".sv": "//",
    ".svh": "//",
    ".tex": "%",
    ".txt": "//",
    ".v": "//",
    ".vh": "//",
}


def test_config():
    """Basic Testing on Config."""
    config = Config()

    assert not config.template_paths
    assert config.existing == Existing.KEEP_TIMESTAMP
    assert config.diffout is None
    assert config.verbose is False
    assert config.inplace_eol_comment is None
    assert config.comment_map == COMMENT_MAP
    assert config.inplace_marker == "GENERATE INPLACE"
    assert config.template_marker == "MAKO TEMPLATE"
    assert config.static_marker == "STATIC"


def test_config_modified():
    """Modified Configuration."""
    config = Config(
        template_paths=[Path("foo"), Path("bar")],
        existing=Existing.KEEP,
        diffout=print,
        verbose=True,
        inplace_marker="INP",
        template_marker="TPL",
        static_marker="STAT",
    )

    assert config.template_paths == [Path("foo"), Path("bar")]
    assert config.existing == Existing.KEEP
    assert config.diffout is print
    assert config.verbose is True
    assert config.inplace_eol_comment is None
    assert config.comment_map == COMMENT_MAP
    assert config.inplace_marker == "INP"
    assert config.template_marker == "TPL"
    assert config.static_marker == "STAT"



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
The Makolator.

A simple API to an improved Mako.
"""

import hashlib
import io
import tempfile
from collections.abc import Generator, Iterator
from contextlib import contextmanager
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryFile

from attrs import define, field
from mako.exceptions import text_error_template
from mako.lookup import TemplateLookup
from mako.runtime import Context
from mako.template import Template
from outputfile import Existing, State, open_
from uniquer import uniquelist

from . import escape, helper
from ._inplace import InplaceRenderer
from ._staticcode import StaticCode, read
from ._util import LOGGER, Paths, humanify, iter_files, norm_paths
from .config import Config
from .datamodel import Datamodel
from .exceptions import MakolatorError
from .info import Info
from .tags import Tag
from .tracker import AddState, Tracker

HELPER = {
    "indent": helper.indent,
    "prefix": helper.prefix,
    "run": helper.run,
    "tex": escape.tex,
}


@define
class Makolator:
    """
    The Makolator.

    A simple API to an improved http://www.makotemplates.org/
    """

    config: Config = field(factory=Config)
    """The Configuration."""

    datamodel: Datamodel = field(factory=Datamodel)
    """The Data Container."""

    info: Info = field(factory=Info)
    """Makolator Information."""

    tracker: Tracker = field(factory=Tracker)
    """File Change Tracker."""

    __cache_path: Path | None = None

    def __del__(self):
        if self.__cache_path:
            rmtree(self.__cache_path)
            self.__cache_path = None

    @property
    def cache_path(self) -> Path:
        """Cache Path."""
        cache_path = self.config.cache_path
        if cache_path:
            cache_path.mkdir(parents=True, exist_ok=True)
            return cache_path

        if not self.__cache_path:
            self.__cache_path = Path(tempfile.mkdtemp(prefix="makolator"))
        return self.__cache_path

    def remove(self, filepaths: Paths):
        """Remove files or files in given directories."""
        for filepath in iter_files(norm_paths(filepaths)):
            self._remove_file(filepath)

    def _remove_file(self, filepath: Path):
        try:
            filepath.unlink()
            self._track_state(filepath, AddState.REMOVED)

        except (PermissionError, FileNotFoundError):
            self._track_state(filepath, State.FAILED)

    @contextmanager
    def open_outputfile(self, filepath: Path, encoding: str = "utf-8", **kwargs):
        """
        Open Outputfile and Return Context.

        Args:
            filepath: path of the created/updated file.

        Keyword Args:
            encoding: Charset.
            kwargs: Additional arguments are forwarded to open.

        Example:

            >>> mklt = Makolator(config=Config(verbose=True))
            >>> with mklt.open_outputfile("myfile.txt") as file:
            ...     file.write("data")
            'myfile.txt'... CREATED.
            >>> with mklt.open_outputfile("myfile.txt") as file:
            ...     file.write("data")
            'myfile.txt'... identical. untouched.
        """
        config = self.config
        kwargs.setdefault("existing", config.existing)
        state = State.FAILED
        try:
            with open_(filepath, encoding=encoding, mkdir=True, diffout=self.config.diffout, **kwargs) as file:
                yield file
            state = file.state
        finally:
            self._track_state(filepath, state)

    def _track_state(self, filepath: Path, state: State) -> None:
        # Track State
        config = self.config
        if config.track:
            self.tracker.add(filepath, state)
        if config.verbose:
            print(f"'{filepath!s}'... {state.value}")

    def gen(self, template_filepaths: Paths, dest: Path | None = None, context: dict | None = None):
        """
        Render template file.

        Args:
            template_filepaths: Templates.

        Keyword Args:
            dest: Output File.
            context: Key-Value Pairs pairs forwarded to the template.
        """
        template_filepaths = norm_paths(template_filepaths)
        LOGGER.debug("_gen(%r, %r)", [str(filepath) for filepath in template_filepaths], str(dest or "STDOUT"))
        is_recursive = any(path.is_dir() for path in template_filepaths)
        if is_recursive:
            self._check_recursive(template_filepaths, dest)
            datamodel = self.datamodel
            for tplbasepath, path in self._iter_recursive(template_filepaths):
                tplpath = tplbasepath / path
                outpath = dest / Template(str(path).removesuffix(".mako")).render(datamodel=datamodel)
                if tplpath.name.endswith(".mako"):
                    self._gen_file([tplpath], outpath, context)
                else:
                    with self.open_outputfile(outpath) as output:
                        output.write(tplpath.read_text())
        else:
            self._gen_file(template_filepaths, dest, context)

    @staticmethod
    def _check_recursive(template_filepaths: list[Path], dest: Path | None = None):
        if dest:
            if not dest.exists() or not dest.is_dir():
                raise ValueError(f"Destination ({str(dest)!r}) must not exists or has to be a directory")
        else:
            raise ValueError("Destination is required")
        if not all(path.is_dir() or not path.exists() for path in template_filepaths):
            raise ValueError("All templates must not exist or have to be a directory")

    @staticmethod
    def _iter_recursive(template_paths: list[Path]) -> Iterator[tuple[Path, Path]]:
        for basepath in template_paths:
            paths = sorted(basepath.glob("**/*"))
            if not paths:
                continue
            for path in paths:
                if path.is_file():
                    yield basepath, path.relative_to(basepath)
            break

    def _gen_file(self, template_filepaths: list[Path], dest: Path | None = None, context: dict | None = None):
        tplfilepaths, lookup = self._create_template_lookup(
            template_filepaths, self.config.template_paths, required=True
        )
        templates = self._create_templates(tplfilepaths, lookup)
        context = context or {}
        comment_sep = self._get_comment_sep(dest)
        if dest is None:
            # newlines may be broken on STDOUT under windows - WON'T FIX
            with TemporaryFile(mode="w+", newline="") as out:
                with read(dest, comment_sep, self.config) as staticcode:
                    template = next(templates)  # Load template
                    LOGGER.info("gen(%r, STDOUT)", template.filename)
                    self._render(template, out, None, context, staticcode, comment_sep)
                out.seek(0)
                for line in out:
                    print(line.rstrip())
        else:
            # Mako takes care about proper newline handling. Therefore we deactivate
            # the universal newline mode, by setting newline="".
            with self.open_outputfile(dest, newline="") as output:
                with read(dest, comment_sep, self.config) as staticcode:
                    template = next(templates)  # Load template
                    LOGGER.info("gen(%r, %r)", template.filename, str(dest))
                    self._render(template, output, dest, context, staticcode, comment_sep)

    def _render(
        self, template: Template, output, dest: Path | None, context: dict, staticcode: StaticCode, comment_sep: str
    ):
        context = Context(output, **self._get_render_context(dest, context, staticcode, comment_sep))
        template.render_context(context)

    def inplace(
        self,
        template_filepaths: Paths,
        filepath: Path,
        context: dict | None = None,
        ignore_unknown: bool = False,
    ):
        """
        Update generated code within `filename` between BEGIN/END markers.

        Args:
            template_filepaths: Templates.
            filepath: File to update.

        Keyword Args:
            context: Key-Value Pairs pairs forwarded to the template.
            ignore_unknown: Ignore unknown inplace markers, instead of raising an error.
        """
        template_filepaths = norm_paths(template_filepaths)
        LOGGER.debug("_inplace(%r, %r)", [str(filepath) for filepath in template_filepaths], str(filepath))
        tplfilepaths, lookup = self._create_template_lookup(template_filepaths, self.config.template_paths)
        templates = tuple(self._create_templates(tplfilepaths, lookup))
        config = self.config
        context = context or {}
        comment_sep = self._get_comment_sep(filepath)

        eol = self._get_eol(filepath, config.inplace_eol_comment)
        inplace = InplaceRenderer(config, templates, ignore_unknown, eol)

        if not filepath.exists() and config.create:
            LOGGER.info("create inplace(%r, %r)", str(tplfilepaths[0]) if tplfilepaths else None, str(filepath))
            self._create_inplace(inplace, filepath, config, comment_sep, context)

        LOGGER.info("inplace(%r, %r)", str(tplfilepaths[0]) if tplfilepaths else None, str(filepath))
        with self.open_outputfile(filepath, existing=Existing.KEEP_TIMESTAMP, newline="") as outputfile:
            with read(filepath, comment_sep, config) as staticcode:
                rendercontext = self._get_render_context(filepath, context, staticcode, comment_sep, inplace=True)
                inplace.render(lookup, filepath, outputfile, rendercontext)

    def _create_inplace(
        self, inplace: InplaceRenderer, filepath: Path, config: Config, comment_sep: str, context: dict
    ):
        func = inplace.get_func("create_inplace")
        if func:
            staticcode = StaticCode.from_config(config, comment_sep)
            rendercontext = self._get_render_context(filepath, context, staticcode, comment_sep, inplace=True)
            buffer = io.StringIO()
            try:
                func.render_context(Context(buffer, **rendercontext))
            except Exception as exc:
                debug = str(text_error_template().render())
                raise MakolatorError(f"Function 'create_inplace' invocation failed. {exc!r}. {debug}") from exc
            with self.open_outputfile(filepath, existing=Existing.KEEP_TIMESTAMP, newline="") as outputfile:
                outputfile.write(buffer.getvalue())
        else:
            raise MakolatorError("None of the templates implements 'create_inplace'")

    def _create_templates(self, tplfilepaths: list[Path], lookup: TemplateLookup) -> Generator[Template, None, None]:
        for tplfilepath in tplfilepaths:
            yield lookup.get_template(tplfilepath.name)
        yield Template(
            """<%! from makolator import helper %>
<%def name="run(*args, **kwargs)">\
${helper.run(*args, **kwargs)}\
</%def>"""
        )

    def _create_template_lookup(
        self, template_filepaths: list[Path], searchpaths: list[Path], required: bool = False
    ) -> tuple[list[Path], TemplateLookup]:
        cache_path = self.cache_path
        tplfilepaths = list(self._find_files(template_filepaths, searchpaths, required=required))
        lookuppaths = uniquelist([tplfilepath.parent for tplfilepath in tplfilepaths] + searchpaths)

        def get_module_filename(filepath: str, uri: str):
            hash_ = hashlib.sha256()
            hash_.update(bytes(filepath, encoding="utf-8"))
            ident = hash_.hexdigest()
            return cache_path / f"{Path(filepath).name}_{ident}.py"

        lookup = TemplateLookup(
            directories=[str(item) for item in lookuppaths],
            cache_dir=self.cache_path,
            input_encoding="utf-8",
            output_encoding="utf-8",
            modulename_callable=get_module_filename,
            strict_undefined=True,
        )
        return tplfilepaths, lookup

    @staticmethod
    def _find_files(
        filepaths: list[Path], searchpaths: list[Path], required: bool = False
    ) -> Generator[Path, None, None]:
        """Find `filepath` in `searchpaths` and return first match."""
        found = False
        for filepath in filepaths:
            if filepath.is_absolute():
                # absolute
                if filepath.exists():
                    yield filepath
                    found = True
            else:
                # relative
                for searchpath in searchpaths:
                    joined = searchpath / filepath
                    if joined.exists():
                        yield joined
                        found = True
        if not found and required:
            if not searchpaths:
                raise MakolatorError(f"None of the templates {humanify(filepaths)}.")
            raise MakolatorError(f"None of the templates {humanify(filepaths)} found at {humanify(searchpaths)}.")

    def _get_render_context(
        self,
        output_filepath: Path | None,
        context: dict,
        staticcode: StaticCode,
        comment_sep: str,
        inplace: bool = False,
    ) -> dict:
        result = dict(context)
        result.update(HELPER)
        tags = [Tag.GENERATED.value]
        if inplace:
            tags.append(Tag.INPLACE_GENERATED.value)
        elif staticcode.is_volatile:
            tags.append(Tag.FULLY_GENERATED.value)
        result["datamodel"] = self.datamodel
        result["makolator"] = self
        result["output_filepath"] = output_filepath
        result["output_tags"] = tuple(tags)
        result["staticcode"] = staticcode
        result["comment"] = helper.prefix(f"{comment_sep} ", rstrip=True)
        return result

    def _get_comment_sep(self, filepath: Path | None, default: str = "//") -> str:
        if not filepath:
            return default
        return self.config.comment_map.get(filepath.suffix, default)

    def _get_eol(self, filepath: Path, eol_comment: str | None):
        if eol_comment:
            sep = self._get_comment_sep(filepath)
            return f"{sep} {eol_comment}"
        return ""

    def clean(self, filepaths: Paths):
        """Remove Fully-Generated Files from Filepaths."""
        for filepath in iter_files(norm_paths(filepaths)):
            is_fully_generated = self.is_fully_generated(filepath)
            if is_fully_generated:
                self._remove_file(filepath)
            elif is_fully_generated is False:
                self._track_state(filepath, State.IDENTICAL)

    def is_fully_generated(self, filepath: Path) -> bool | None:
        """Check If File Is Fully Generated."""
        try:
            with filepath.open("r") as file:
                for _, line in zip(range(self.config.tag_lines), file, strict=False):
                    if Tag.FULLY_GENERATED.value in line:
                        return True
            return False
        except UnicodeDecodeError:  # binary files
            return None



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Static Code Preservation."""

import os
import re
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

from attrs import define, field

from ._util import LOGGER, check_indent, fill_marker, humanify
from .config import Config
from .exceptions import MakolatorError

StaticCodeMap = dict[str, str]


@define
class Info:
    """Static Code Context Information."""

    lineno: int
    indent: str
    name: str


@define
class StaticCode:
    """Static Code Manager."""

    comment_sep: str
    staticcodemap: StaticCodeMap
    marker: str
    marker_fill: str = ""
    marker_linelength: int = 0

    _names: list[str] = field(factory=list)

    def __call__(self, name, default=None, comment_sep=None):
        if name in self._names:
            raise MakolatorError(f"duplicate static code {name!r}")
        self._names.append(name)
        if comment_sep is None:
            comment_sep = self.comment_sep or ""
        code = self.staticcodemap.pop(name, default) or ""
        cpre = f"{comment_sep} " if comment_sep else ""
        begin = f"{cpre}{self.marker} BEGIN {name}"
        end = f"{cpre}{self.marker} END {name}"
        if self.marker_fill and self.marker_linelength:
            begin = fill_marker(begin, self.marker_fill, self.marker_linelength)
            end = fill_marker(end, self.marker_fill, self.marker_linelength)
        lines = [begin]
        lines.extend(code.splitlines())
        lines.append(end)
        return os.linesep.join(lines)

    @property
    def is_volatile(self) -> bool:
        """Static Code is Empty."""
        return not self.staticcodemap

    @staticmethod
    def from_config(config: Config, comment_sep: str, staticcodemap: StaticCodeMap | None = None) -> "StaticCode":
        staticcodemap = staticcodemap or {}
        return StaticCode(
            comment_sep=comment_sep,
            staticcodemap=staticcodemap,
            marker=config.static_marker,
            marker_fill=config.marker_fill,
            marker_linelength=config.marker_linelength,
        )


@contextmanager
def read(filepath: Path | None, comment_sep: str, config: Config) -> Iterator[StaticCode]:
    """Read from ``filepath``."""
    staticcodemap: StaticCodeMap = {}
    _read(filepath, config.static_marker, staticcodemap)
    yield StaticCode.from_config(config, comment_sep, staticcodemap=staticcodemap)
    if staticcodemap:
        names = humanify(staticcodemap)
        raise MakolatorError(f"'{filepath!s}': unknown static code {names}")


def _read(filepath: Path | None, marker: str, staticcodemap: StaticCodeMap):
    if filepath and marker:
        begin = re.compile(rf"(?P<indent>\s*).*{marker}\s+BEGIN\s+(?P<name>\S+)\s*")
        info = None

        try:
            with filepath.open(encoding="utf-8", newline="") as file:
                fileiter = enumerate(file, 1)
                while True:
                    if info:
                        # process static code
                        _process(filepath, marker, staticcodemap, fileiter, begin, info)
                        info = None
                    else:
                        # normal lines
                        while True:
                            lineno, line = next(fileiter)
                            # search for BEGIN
                            beginmatch = begin.match(line)
                            if beginmatch:
                                # consume BEGIN
                                info = Info(lineno, **beginmatch.groupdict())
                                break

        except (StopIteration, FileNotFoundError):
            pass
        if info:
            raise MakolatorError(f"'{filepath!s}:{info.lineno}' BEGIN without END.")


def _process(filepath: Path, marker: str, staticcodemap: StaticCodeMap, fileiter, begin, info: Info):
    end = re.compile(rf"(?P<indent>\s*).*{marker}\s+END\s+(?P<name>\S+)\s*")
    lines: list[str] = []
    while True:
        # search END
        lineno, line = next(fileiter)

        beginmatch = begin.match(line)
        if beginmatch:
            msg = f"missing END tag {info.name!r} for '{filepath!s}:{info.lineno}'"
            raise MakolatorError(msg)

        endmatch = end.match(line)
        if endmatch:
            # consume END
            LOGGER.debug("Static Code %r at '%s:%d'", info.name, str(filepath), info.lineno)
            if info.name not in staticcodemap:
                staticcodemap[info.name] = "".join(lines)
            else:
                msg = f"duplicate static code {info.name!r} at '{filepath!s}:{info.lineno}'"
                raise MakolatorError(msg)
            check_indent(filepath, lineno, info.indent, endmatch.group("indent"))
            break
        lines.append(line)



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Update Tracker."""

from enum import Enum
from pathlib import Path
from typing import TypeAlias

from attrs import define, field
from outputfile import State


class AddState(Enum):
    """Additional States."""

    REMOVED = "REMOVED."


_STAT_INIT = {
    State.UPDATED: 0,
    State.IDENTICAL: 0,
    State.CREATED: 0,
    State.OVERWRITTEN: 0,
    State.EXISTING: 0,
    State.FAILED: 0,
    AddState.REMOVED: 0,
}

FileState: TypeAlias = State | AddState


@define
class Tracker:
    """Update Tracker."""

    _items: list[tuple[Path, FileState]] = field(factory=list)
    _stat: dict[FileState, int] = field(factory=lambda: dict(_STAT_INIT))

    def add(self, path: Path, state: FileState) -> None:
        """Add Information."""
        self._items.append((path, state))
        self._stat[state] += 1

    def clear(self):
        """Clear Information."""
        self._items.clear()
        self._stat = dict(_STAT_INIT)

    @property
    def total(self):
        """Total Number of Files."""
        return sum(self._stat.values())

    @property
    def stat(self) -> str:
        """Status Summary."""
        counts = (f"{count} {state.value}" for state, count in self._stat.items() if count)
        return " ".join((f"{self.total} files.", *counts))

    @property
    def updated(self) -> int:
        """Files Updated."""
        return self._stat[State.UPDATED]

    @property
    def identical(self) -> int:
        """Files Identical."""
        return self._stat[State.IDENTICAL]

    @property
    def created(self) -> int:
        """Files Created."""
        return self._stat[State.CREATED]

    @property
    def overwritten(self) -> int:
        """Files Overwritten."""
        return self._stat[State.OVERWRITTEN]

    @property
    def existing(self) -> int:
        """Files Existing."""
        return self._stat[State.EXISTING]

    @property
    def failed(self) -> int:
        """Files Failed."""
        return self._stat[State.FAILED]

    @property
    def removed(self) -> int:
        """Files Removed."""
        return self._stat["rm"]



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Configuration Handling."""

from collections.abc import Callable
from pathlib import Path

from attrs import define
from outputfile import Existing

COMMENT_MAP_DEFAULT = {
    ".c": "//",
    ".c++": "//",
    ".cpp": "//",
    ".ini": "#",
    ".py": "#",
    ".sv": "//",
    ".svh": "//",
    ".tex": "%",
    ".txt": "//",
    ".v": "//",
    ".vh": "//",
}


@define
class Config:
    """
    Configuration.

    Container For All Customization Options.

    The following file extensions are known:

        >>> for suffix, comment in Config().comment_map.items():
        ...     print(f"{suffix}: {comment}")
        .c: //
        .c++: //
        .cpp: //
        .ini: #
        .py: #
        .sv: //
        .svh: //
        .tex: %
        .txt: //
        .v: //
        .vh: //
    """

    template_paths: list[Path] = []  # noqa: RUF008
    """Default Search Paths for Templates."""

    existing: Existing = Existing.KEEP_TIMESTAMP
    """Behaviour in case of existing files."""

    diffout: Callable[[str], None] | None = None
    """``print`` function to handle differential output on changed files."""

    verbose: bool = False
    """Enable Verbose Output."""

    create: bool = False
    """Create Missing Inplace Files."""

    template_marker: str = "MAKO TEMPLATE"
    """Search marker for template code within output file."""

    inplace_marker: str = "GENERATE INPLACE"
    """Search marker for output code within output file."""

    static_marker: str = "STATIC"
    """Marker for Static Code."""

    marker_fill: str = "="
    """Marker Filling."""

    marker_linelength: int = 0
    """Marker Line Length for Filling."""

    cache_path: Path | None = None
    """
    Cache Directory.

    Used to store converted templates. Use if you have many and/or large templates.
    Speeds up rendering. Share it between runs.
    """

    comment_map: dict[str, str] = COMMENT_MAP_DEFAULT
    """
    Line Comment Symbols.

    File Suffix dependent comment starter.
    """

    inplace_eol_comment: str | None = None
    """End-Of-Line Comment Added On Every Inplace Generated Line """

    track: bool = False
    """Track Changes."""

    tag_lines: int = 50
    """Number Of Lines Within A Files to look for Tags."""



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Inplace Generation."""

import io
import os
import re
from pathlib import Path
from typing import Any

from aligntext import Align
from attrs import define, field
from mako.exceptions import text_error_template
from mako.lookup import TemplateLookup
from mako.runtime import Context
from mako.template import Template

from ._util import LOGGER, check_indent, fill_marker
from .config import Config
from .exceptions import MakolatorError


@define
class InplaceInfo:
    """Inplace Rendering Context Information."""

    lineno: int
    indent: str
    funcname: str
    args: str
    func: Any
    end: Any


@define
class TplInfo:
    """Template Context Information."""

    lineno: int
    pre: str
    lines: list[str] = field(factory=list)


@define
class InplaceRenderer:
    """Inplace Renderer."""

    config: Config
    templates: tuple[Template, ...]
    ignore_unknown: bool
    eol: str

    def render(self, lookup: TemplateLookup, filepath: Path, outputfile, context: dict):  # noqa: C901
        """Render."""
        inplace_marker = self.config.inplace_marker
        ibegin = re.compile(rf"(?P<indent>\s*).*{inplace_marker}\s+BEGIN\s(?P<funcname>[a-z_]+)\((?P<args>.*)\)")
        iinfo = None

        template_marker = self.config.template_marker
        tinfo = None
        tbegin = re.compile(rf"(?P<pre>.*)\s*{template_marker}\s+BEGIN")
        templates = list(self.templates)

        with filepath.open(encoding="utf-8", newline="") as inputfile:
            inputiter = enumerate(inputfile.readlines(), 1)
            try:
                while True:
                    if iinfo:
                        # GENERATE INPLACE
                        self._process_inplace(filepath, outputfile, context, inputiter, iinfo, ibegin)
                        iinfo = None

                    elif tinfo:
                        # MAKO TEMPLATE
                        self._process_template(filepath, lookup, outputfile, templates, inputiter, tinfo, tbegin)
                        tinfo = None

                    else:
                        # normal lines
                        while True:
                            lineno, line = next(inputiter)
                            if inplace_marker:
                                # search for "INPLACE BEGIN <funcname>(<args>)"
                                beginmatch = ibegin.match(line)
                                if beginmatch:
                                    outputfile.write(self._fill_marker(beginmatch))
                                    # consume INPLACE BEGIN
                                    iinfo = self._start_inplace(templates, filepath, lineno, **beginmatch.groupdict())
                                    break
                            if template_marker:
                                # search for "TEMPLATE BEGIN"
                                beginmatch = tbegin.match(line)
                                if beginmatch:
                                    outputfile.write(self._fill_marker(beginmatch))
                                    # consume TEMPLATE BEGIN
                                    tinfo = TplInfo(lineno, beginmatch.group("pre"))
                                    break
                            outputfile.write(line)

            except StopIteration:
                pass
        if iinfo:
            raise MakolatorError(f"'{filepath!s}:{iinfo.lineno}' BEGIN {iinfo.funcname}({iinfo.args}) without END.")
        if tinfo:
            raise MakolatorError(f"'{filepath!s}:{tinfo.lineno}' BEGIN without END.")

    def _process_inplace(self, filepath: Path, outputfile, context: dict, inputiter, iinfo, ibegin):
        while True:
            # search for "INPLACE END"
            lineno, line = next(inputiter)

            beginmatch = ibegin.match(line)
            if beginmatch:
                msg = f"missing END tag {iinfo.funcname!r} for '{filepath!s}:{iinfo.lineno}'"
                raise MakolatorError(msg)

            endmatch = iinfo.end.match(line)
            if endmatch:
                # fill
                self._fill_inplace(filepath, outputfile, iinfo, context)
                # propagate INPLACE END tag
                outputfile.write(self._fill_marker(endmatch))
                check_indent(filepath, lineno, iinfo.indent, endmatch.group("indent"))
                # consume INPLACE END
                break

    def _process_template(
        self, filepath: Path, lookup: TemplateLookup, outputfile, templates, inputiter, tinfo, tbegin
    ):
        # capture TEMPLATE
        pre = tinfo.pre
        prelen = len(pre)
        tend = re.compile(rf"(?P<pre>.*)\s*{self.config.template_marker}\s+END")
        while True:
            _, line = next(inputiter)

            beginmatch = tbegin.match(line)
            if beginmatch:
                msg = f"missing END tag for '{filepath!s}:{tinfo.lineno}'"
                raise MakolatorError(msg)

            # search for "INPLACE END"
            endmatch = tend.match(line)
            if endmatch:
                outputfile.write(self._fill_marker(endmatch))
                LOGGER.debug("Template '%s:%d'", str(outputfile), tinfo.lineno)
                templates.append(Template("".join(tinfo.lines), lookup=lookup))
                break
            # propagate
            outputfile.write(line)

            if line.startswith(pre):
                line = line[prelen:]
            tinfo.lines.append(line)

    def get_func(self, funcname: str):
        """Retrieve `funcname` from templates."""
        return self._get_func(list(self.templates), funcname)

    @staticmethod
    def _get_func(templates: list[Template], funcname: str):
        for template in templates:
            try:
                return template.get_def(funcname)
            except AttributeError:  # noqa: PERF203
                continue
        return None

    def _start_inplace(
        self, templates: list[Template], filepath: Path, lineno: int, indent: str, funcname: str, args: str
    ) -> InplaceInfo | None:
        func = self._get_func(templates, funcname)
        if func:
            end = re.compile(rf"(?P<indent>\s*).*{self.config.inplace_marker}\s+END\s{funcname}")
            return InplaceInfo(lineno, indent, funcname, args, func, end)
        if not self.ignore_unknown:
            raise MakolatorError(f"{filepath!s}:{lineno} Function '{funcname}' is not found in templates.")
        return None

    def _fill_inplace(self, filepath: Path, outputfile, inplace: InplaceInfo, context: dict):
        LOGGER.debug("Inplace '%s:%d'", str(filepath), inplace.lineno)
        # determine args, kwargs
        try:
            args, kwargs = eval(f"_extract({inplace.args})", {"_extract": _extract})  # noqa: S307
        except Exception as exc:
            raise MakolatorError(
                f"{filepath!s}:{inplace.lineno} Function invocation failed. "
                f"{exc!r} in arguments: '{inplace.funcname}({inplace.args})'."
            ) from exc

        # run func(args, kwargs)
        buffer = io.StringIO()
        indent = inplace.indent
        context = Context(buffer, **context)
        try:
            inplace.func.render_context(context, *args, **kwargs)
        except Exception as exc:
            debug = str(text_error_template().render())
            raise MakolatorError(
                f"{filepath!s}:{inplace.lineno} Function '{inplace.funcname}' invocation failed. {exc!r}. {debug}"
            ) from exc
        eol = self.eol
        lines = buffer.getvalue().splitlines()
        linesep = os.linesep
        if eol:
            align = Align()
            for line in lines:
                align.add_row(line, eol)
            for item in align:
                outputfile.write(f"{indent}{item}{linesep}")
        else:
            for line in lines:
                if line:
                    outputfile.write(f"{indent}{line}{linesep}")
                else:
                    outputfile.write(linesep)

        buffer.close()

    def _fill_marker(self, mat: re.Match) -> str:
        marker_fill = self.config.marker_fill
        marker_linelength = self.config.marker_linelength
        if marker_fill and marker_linelength:
            line = mat.string[mat.start() : mat.end()]
            return fill_marker(line, marker_fill, marker_linelength) + os.linesep
        return mat.string


def _extract(*args, **kwargs):
    return (args, kwargs)



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Makolator Helper."""

import os
import subprocess
import tempfile
from typing import Any


def run(args, **kwargs):
    """
    Run External Command And Use Command STDOUT as Result.

    :any:`subprocess.run` wrapper.
    STDOUT is taken as result.

    The variable ``${TMPDIR}`` in the arguments will be replaced by a temporary directory
    path.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        if isinstance(args, str):
            args = args.replace("${TMPDIR}", tmpdir)
        else:
            args = [arg.replace("${TMPDIR}", tmpdir) for arg in args]
        kwargs["stdout"] = subprocess.PIPE
        result = subprocess.run(args, check=True, **kwargs)  # noqa: S603
        return result.stdout.decode("utf-8").rstrip()


def indent(text_or_int: Any, rstrip: bool = False):
    """
    Indent Lines by number of ``text_or_int``.

        >>> print(indent('''A # doctest: +SKIP
        ... B
        ... C'''))
          A
          B
          C

        >>> print(indent(4)('''A # doctest: +SKIP
        ... B
        ... C'''))
            A
            B
            C
    """
    if isinstance(text_or_int, int):
        return prefix(" " * text_or_int, rstrip=rstrip)
    return prefix("  ", rstrip=rstrip)(text_or_int)


def prefix(pre: str, rstrip: bool = False):
    """
    Add ``pre`` In Front of Every Line.

        >>> print(prefix('PRE-')('''A # doctest: +SKIP
        ... B
        ... C'''))
        PRE-A
        PRE-B
        PRE-C
    """
    if rstrip:

        def func(text):
            return os.linesep.join(f"{pre}{line}".rstrip() for line in text.splitlines())

    else:

        def func(text):
            return os.linesep.join(f"{pre}{line}" for line in text.splitlines())

    return func



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Command Line Interface.
"""

import argparse
from pathlib import Path

from makolator import Config, Existing, Info, Makolator, get_cli


def main(args=None):
    """Command Line Interface Processing."""
    parser = argparse.ArgumentParser(
        prog="makolator",
        description="Mako Templates (https://www.makotemplates.org/) extended.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="cmd")
    default_config = Config()

    gen = subparsers.add_parser(
        "gen",
        help="Generate File",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""\
Generate a file from a template:

    makolator gen test.txt.mako test.txt

Generate a file from a template and fallback to 'default.txt.mako' if 'test.txt.mako' is missing:

    makolator gen test.txt.mako default.txt.mako test.txt
""",
    )
    gen.add_argument("templates", nargs="+", type=Path, help="Template Files. At least one must exist")
    gen.add_argument("output", type=Path, help="Output File")

    inplace = subparsers.add_parser(
        "inplace",
        help="Update File Inplace",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""\
Update with inplace template only:

    makolator inplace test.txt

Update a file from a template:

    makolator inplace test.txt.mako test.txt

Update a file from a template and fallback to 'default.txt.mako' if 'test.txt.mako' is missing:

    makolator inplace test.txt.mako default.txt.mako test.txt
""",
    )
    inplace.add_argument("templates", nargs="*", type=Path, help="Optional Template Files")
    inplace.add_argument("inplace", type=Path, help="Updated File")
    inplace.add_argument(
        "--ignore-unknown",
        "-i",
        action="store_true",
        help="Ignore unknown template function calls.",
    )

    clean = subparsers.add_parser(
        "clean",
        help="Remove Fully-Generated Files",
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="""\
Remove all files with '@fully-generated' in header.
The number of inspected lines at the top of a file is defined by --tag_lines.

    makolator clean .

""",
    )
    clean.add_argument("paths", nargs="+", type=Path, help="Paths to look for files.")

    for sub in (gen, inplace, clean):
        sub.add_argument("--verbose", "-v", action="store_true", help="Tell what happens to the file.")
        sub.add_argument("--show-diff", "-s", action="store_true", help="Show what lines changed.")
        sub.add_argument(
            "--tag_lines",
            default=default_config.tag_lines,
            help=f"Number of Inspected Lines on 'clean'. Default is {default_config.tag_lines}.",
        )
        sub.add_argument("--stat", "-S", action="store_true", help="Print Statistics")
    for sub in (gen, inplace):
        sub.add_argument(
            "--existing",
            "-e",
            default="keep_timestamp",
            choices=[item.value for item in Existing],
            help="What if the file exists. Default is 'keep_timestamp'",
        )
        sub.add_argument(
            "--template-path",
            "-T",
            type=Path,
            default=[],
            action="append",
            help="Directories with templates referred by include/inherit/...",
        )
        sub.add_argument(
            "--marker-fill",
            type=str,
            help=(
                "Static Code, Inplace and Template Marker are filled with "
                "this given value until reaching line length of --marker-linelength."
            ),
        )
        sub.add_argument(
            "--marker-linelength",
            type=int,
            help=("Static Code, Inplace and Template Marker are filled until --marker-linelength."),
        )
        sub.add_argument("--eol", "-E", help="EOL comment on generated lines")
        sub.add_argument("--create", "-c", action="store_true", default=False, help="Create Missing Inplace File")

    args = parser.parse_args(args=args)
    if args.cmd:
        if args.cmd == "clean":
            config = Config(
                verbose=args.verbose,
                diffout=print if args.show_diff else None,
                tag_lines=args.tag_lines,
                track=args.stat,
            )
        else:
            config = Config(
                verbose=args.verbose,
                create=args.create,
                diffout=print if args.show_diff else None,
                existing=args.existing,
                template_paths=[*args.template_path, Path()],
                marker_fill=args.marker_fill,
                marker_linelength=args.marker_linelength,
                inplace_eol_comment=args.eol,
                tag_lines=args.tag_lines,
                track=args.stat,
            )
        info = Info(cli=get_cli())
        mklt = Makolator(config=config, info=info)
        if args.cmd == "gen":
            mklt.gen(args.templates, args.output)
        elif args.cmd == "inplace":
            mklt.inplace(args.templates, args.inplace, ignore_unknown=args.ignore_unknown)
        elif args.cmd == "clean":
            mklt.clean(args.paths)
        if config.track:
            print(mklt.tracker.stat)
    else:
        parser.print_help()



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Data Model."""


class Datamodel:
    """
    Datamodel.

    A simple container for all data attributes.
    Add attributes on your needs. That's it.

        >>> Datamodel()
        Datamodel()
        >>> Datamodel(abc='def', item=4)
        Datamodel(abc='def', item=4)
        >>> datamodel = Datamodel(abc='def')
        >>> datamodel
        Datamodel(abc='def')
        >>> datamodel.item=4
        >>> datamodel
        Datamodel(abc='def', item=4)
    """

    def __init__(self, **kwargs):
        """Datamodel."""
        self.__dict__.update(kwargs)

    def __repr__(self):
        kwargs = ", ".join(f"{key}={value!r}" for key, value in self.__dict__.items())
        return f"{self.__class__.__name__}({kwargs})"

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            return self.__dict__ == other.__dict__
        return NotImplemented



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Datamodel Testing."""

import re
from pathlib import Path
from shutil import copyfile

from pytest import raises
from test2ref import assert_refdata

from makolator import MakolatorError
from makolator.cli import main

FILEPATH = Path(__file__)
TESTDATA = FILEPATH.parent / "testdata"
REFDATA = FILEPATH.parent / "refdata" / FILEPATH.stem


def test_gen(tmp_path, capsys):
    """Gen."""
    main(["gen", str(TESTDATA / "test.txt.mako"), str(tmp_path / "test.txt")])
    assert_refdata(test_gen, tmp_path, capsys=capsys)


def test_gen_stat(tmp_path, capsys):
    """Gen."""
    main(["gen", str(TESTDATA / "test.txt.mako"), str(tmp_path / "test.txt"), "--stat"])
    assert_refdata(test_gen_stat, tmp_path, capsys=capsys)


def test_inplace(tmp_path, capsys):
    """Inplace."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace.txt", filepath)
    main(["inplace", str(TESTDATA / "inplace.txt.mako"), str(filepath)])
    assert_refdata(test_inplace, tmp_path, capsys=capsys)


def test_inplace_stat(tmp_path, capsys):
    """Inplace."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace.txt", filepath)
    main(["inplace", "--stat", str(TESTDATA / "inplace.txt.mako"), str(filepath)])
    assert_refdata(test_inplace_stat, tmp_path, capsys=capsys)


def test_inplace_missing(tmp_path):
    """Inplace File Not Found."""
    with raises(FileNotFoundError):
        main(["inplace", str(TESTDATA / "inplace.txt.mako"), str(tmp_path / "inplace.txt")])


def test_inplace_create_missing(tmp_path, caplog):
    """Inplace With Create - missing create_inplace."""
    with raises(MakolatorError, match=re.escape("None of the templates implements 'create_inplace'")):
        main(["inplace", "--create", str(TESTDATA / "inplace.txt.mako"), str(tmp_path / "inplace.txt")])


def test_inplace_create_broken(tmp_path, caplog):
    """Inplace With Create - broken create_inplace."""
    with raises(MakolatorError, match=re.escape("assert False")):
        main(["inplace", "--create", str(TESTDATA / "inplace-create-broken.txt.mako"), str(tmp_path / "inplace.txt")])


def test_inplace_create(tmp_path, caplog):
    """Inplace With Create."""
    main(["inplace", "--create", str(TESTDATA / "inplace-create.txt.mako"), str(tmp_path / "inplace.txt")])
    assert_refdata(test_inplace_create, tmp_path, caplog=caplog)



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
r"""
Extended Mako Templates for Python.

Makolator is not makulation. It extends the [mako template engine](https://www.makotemplates.org/) by the following
features:

* Simple API
* Keep timestamp of generated files, if content did not change (a gift for every build system user)
* Easy hierarchical template usage
* Inplace File Rendering

This is how to use it

## Initialize

Just create your instance of [`Makolator`](#makolator.Makolator)

    >>> from makolator import Makolator
    >>> mklt = Makolator()

## Configure

The config attribute contains all settings.
See [`Config`](#makolator.Config) for a complete documentation.
The most relevant settings are search paths for general templates ...

    >>> mklt.config.template_paths
    []

... and the replacement strategy for `outputfile.Existing` files:

    >>> mklt.config.existing
    <Existing.KEEP_TIMESTAMP: 'keep_timestamp'>

If you want to place templates in the actual working directory - set it as search path:

    >>> from pathlib import Path
    >>> mklt.config.template_paths = [Path('.')]

If you like it verbose - try:

    >>> mklt.config.verbose = True

## Rendering

Assume you have this template in a file
(please ignore the `...` here - they are just there for a proper python example):

    >>> Path('file.txt.mako').write_text('''\\
    ... <%def name="top(name)">\\
    ... generated-top: ${name}
    ... </%def>\\
    ... <%def name="bot(name, **kwargs)">\\
    ... generated-bot: ${name} ${kwargs}
    ... </%def>\\
    ... ${datamodel}
    ... ${top('foo')}\\
    ... ${bot('foo', b=4)}\\
    ... ''') and None

`gen` uses ``sys.stdout`` by default ...

    >>> mklt.gen([Path('file.txt.mako')])
    Datamodel()
    generated-top: foo
    generated-bot: foo {'b': 4}

... or use ``dest`` for file creation - the verbose mode shares some nice information:

    >>> mklt.gen([Path('file.txt.mako')], dest=Path("file.txt"))
    'file.txt'... CREATED.
    >>> mklt.gen([Path('file.txt.mako')], dest=Path("file.txt"))
    'file.txt'... identical. untouched.

## Datamodel

The [`Datamodel`](#makolator.Datamodel) is your container to forward data to the template.

    >>> mklt.datamodel.mydata = {'a': 0, 'b': 1}

The output looks like that:

    >>> mklt.gen([Path('file.txt.mako')])
    Datamodel(mydata={'a': 0, 'b': 1})
    generated-top: foo
    generated-bot: foo {'b': 4}

    and you get notified about the changed content:

    >>> mklt.gen([Path('file.txt.mako')], dest=Path("file.txt"))
    'file.txt'... UPDATED.

## Differential output

If you like it even more verbose - try:

    >>> mklt.config.diffout = print

This will send any diff of the updated files to stdout.

    >>> mklt.datamodel.mydata['b'] = 2
    >>> mklt.gen([Path('file.txt.mako')], dest=Path("file.txt")) # doctest: +SKIP
    ---
    +++
    @@ -1,3 +1,3 @@
    -Datamodel(mydata={'a': 0, 'b': 1})
    +Datamodel(mydata={'a': 0, 'b': 2})
    generated-top: foo
    generated-bot: foo {'b': 4}
    <BLANKLINE>
    'file.txt'... UPDATED.

By default this is disabled:

    >>> mklt.config.diffout = None

## Inplace Rendering

One key feature is the inplace rendering.

Assume you have a handwritten file and you just want to update/generate a part of it.
`GENERATE INPLACE` will become your new friend:

    >>> Path('inplace.txt').write_text('''\\
    ... I am a regular text file.
    ... This is handwritten text.
    ... GENERATE INPLACE BEGIN top('bar')
    ...   this line will be replaced
    ... GENERATE INPLACE END top
    ... This is handwritten text too.
    ... ''') and None

    >>> mklt.inplace([Path('file.txt.mako')], Path("inplace.txt"))
    'inplace.txt'... UPDATED.

    >>> print(Path('inplace.txt').read_text())
    \
    I am a regular text file.
    This is handwritten text.
    GENERATE INPLACE BEGIN top('bar')
    generated-top: bar
    GENERATE INPLACE END top
    This is handwritten text too.
    <BLANKLINE>

If you like to warn your users about generated sections even more, you can add a comment by default

    >>> mklt.config.inplace_eol_comment = "GENERATED"
    >>> mklt.inplace([Path('file.txt.mako')], Path("inplace.txt"))
    'inplace.txt'... UPDATED.

    >>> print(Path('inplace.txt').read_text())
    \
    I am a regular text file.
    This is handwritten text.
    GENERATE INPLACE BEGIN top('bar')
    generated-top: bar // GENERATED
    GENERATE INPLACE END top
    This is handwritten text too.
    <BLANKLINE>

That's it.
"""

from outputfile import Existing

from .config import Config
from .datamodel import Datamodel
from .escape import tex
from .exceptions import MakolatorError
from .helper import indent, prefix, run
from .info import Info, get_cli
from .makolator import Makolator
from .tracker import Tracker

__all__ = [
    "Config",
    "Datamodel",
    "Existing",
    "Info",
    "Makolator",
    "MakolatorError",
    "Tracker",
    "get_cli",
    "indent",
    "prefix",
    "run",
    "tex",
]



#
# MIT License
#
# Copyright (c) 2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Info Testing."""

from pytest import fixture

from makolator import Makolator
from makolator._util import iter_files


@fixture
def mklt():
    """Makolator."""
    mklt = Makolator()
    mklt.config.track = True
    mklt.config.verbose = True
    return mklt


@fixture
def example(tmp_path):
    """Example File Structure."""
    onedir = tmp_path / "one"
    onedir.mkdir()
    twodir = onedir / "two"
    twodir.mkdir()
    # File without @generated tag
    (onedir / "one.txt").write_text("""
Hello World
""")
    # Two generated files
    (tmp_path / "sec.txt").write_text("""
// @fully-generated
""")
    (tmp_path / "third.txt").write_text("""
// @generated
""")
    # @fully-generated tag out of reach
    with (twodir / "four.txt").open("w") as file:
        for _ in range(100):
            file.write("@fully-generated")
    (onedir / "file.bin").write_bytes(bytes(range(256)))
    yield tmp_path


def test_find_files(example):
    """Find Files."""
    assert tuple(iter_files((example,))) == (
        example / "one" / "file.bin",
        example / "one" / "one.txt",
        example / "one" / "two" / "four.txt",
        example / "sec.txt",
        example / "third.txt",
    )


def test_clean(example, mklt, capsys):
    """Clean Operation."""
    onefile = example / "one" / "one.txt"
    secfile = example / "sec.txt"
    thirdfile = example / "third.txt"
    fourfile = example / "one" / "two" / "four.txt"

    mklt.clean(example)

    assert onefile.exists()
    assert not secfile.exists()
    assert thirdfile.exists()
    assert not fourfile.exists()

    assert mklt.tracker.stat == "4 files. 2 identical. untouched. 2 REMOVED."
    assert capsys.readouterr().out.splitlines() == [
        f"'{onefile}'... identical. untouched.",
        f"'{fourfile}'... REMOVED.",
        f"'{secfile}'... REMOVED.",
        f"'{thirdfile}'... identical. untouched.",
    ]


def test_remove_files(example, mklt, capsys):
    """Remove Files."""
    onefile = example / "one" / "one.txt"
    secfile = example / "sec.txt"
    thirdfile = example / "third.txt"
    fourfile = example / "one" / "two" / "four.txt"

    mklt.remove(onefile)

    assert not onefile.exists()
    assert secfile.exists()
    assert thirdfile.exists()
    assert fourfile.exists()

    assert mklt.tracker.stat == "1 files. 1 REMOVED."
    assert capsys.readouterr().out.splitlines() == [f"'{onefile}'... REMOVED."]


def test_remove(example, mklt, capsys):
    """Remove Directories."""
    binfile = example / "one" / "file.bin"
    onefile = example / "one" / "one.txt"
    secfile = example / "sec.txt"
    thirdfile = example / "third.txt"
    fourfile = example / "one" / "two" / "four.txt"

    mklt.remove(example / "one")

    assert not onefile.exists()
    assert secfile.exists()
    assert thirdfile.exists()
    assert not fourfile.exists()

    assert mklt.tracker.stat == "3 files. 3 REMOVED."
    assert capsys.readouterr().out.splitlines() == [
        f"'{binfile}'... REMOVED.",
        f"'{onefile}'... REMOVED.",
        f"'{fourfile}'... REMOVED.",
    ]


def test_remove_failed(example, mklt, capsys):
    """Remove Failed."""
    onefile = example / "one" / "one.txt"

    onefile.unlink()
    mklt._remove_file(onefile)

    assert mklt.tracker.stat == "1 files. 1 FAILED."
    assert capsys.readouterr().out.splitlines() == [f"'{onefile!s}'... FAILED."]



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
Utilities.
"""

import logging
from collections.abc import Iterable
from pathlib import Path
from typing import TypeAlias

Paths: TypeAlias = Path | Iterable[Path]
LOGGER = logging.getLogger("makolator")


def norm_paths(paths: Paths) -> list[Path]:
    """Normalize Single Path or List of Paths to List of Paths."""
    try:
        return list(paths)  # type: ignore[arg-type]
    except TypeError:
        return [paths]  # type: ignore[list-item]


def iter_files(paths: Iterable[Path]) -> Iterable[Path]:
    for path in paths:
        if path.is_file():
            yield path
        elif path.is_dir():
            yield from iter_files(sorted(path.iterdir()))


def check_indent(filepath: Path, lineno: int, beginindent, endindent):
    """Check ``BEGIN``/``END`` indent."""
    if endindent != beginindent:
        LOGGER.warning(
            "'%s:%d' Indent of END tag %r does not match indent of BEGIN tag %r.",
            filepath,
            lineno,
            endindent,
            beginindent,
        )


def humanify(iterable):
    """Just join names from `iterable`."""
    if iterable:
        return ", ".join(repr(str(item)) for item in iterable)
    return "''"


def fill_marker(line: str, fill: str, length: int):
    filllen = len(fill)
    linelen = len(line) + 1
    if linelen >= length:
        return line
    count = (length - linelen) // filllen
    filling = fill * count
    return f"{line} {filling}"



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Makolator Testing."""

import re
from pathlib import Path
from shutil import copyfile

from mako.exceptions import CompileException
from pytest import fixture, raises
from test2ref import assert_paths, assert_refdata

from makolator import Makolator, MakolatorError

FILEPATH = Path(__file__)
TESTDATA = FILEPATH.parent / "testdata"
REFDATA = FILEPATH.parent / "refdata" / FILEPATH.stem


@fixture
def mklt():
    """Default :any:`Makolator` instance with proper ``template_paths``."""
    mklt = Makolator()
    mklt.config.template_paths = [TESTDATA]
    yield mklt


def test_inplace(tmp_path, capsys, caplog):
    """Render File Inplace."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_inplace, tmp_path, capsys=capsys, caplog=caplog)


def test_disabled(tmp_path, capsys, caplog):
    """Render File Inplace."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace.txt", filepath)
    mklt = Makolator()
    mklt.config.template_marker = None
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_disabled, tmp_path, capsys=capsys, caplog=caplog)


def test_indent(tmp_path, caplog):
    """Render File Inplace Indent."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-indent.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_indent, tmp_path, caplog=caplog)


def test_broken_arg(tmp_path):
    """Rarger File Inplace with broken arg."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-broken-arg.txt", filepath)
    mklt = Makolator()
    with raises(MakolatorError, match=re.escape(r"SyntaxError")):
        mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)


def test_broken_end(tmp_path):
    """Render File Inplace with broken end."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-broken-end.txt", filepath)
    mklt = Makolator()
    with raises(MakolatorError, match=re.escape(r"without END.")):
        mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)


def test_broken_func(tmp_path):
    """Render File Inplace with broken gen."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-broken-gen.txt", filepath)
    mklt = Makolator()
    with raises(MakolatorError, match=re.escape(r"ZeroDivisionError: division by zero")):
        mklt.inplace([TESTDATA / "inplace-broken-func.txt.mako"], filepath)


def test_unknown(tmp_path):
    """Render File Inplace with missing func."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-unknown.txt", filepath)
    mklt = Makolator()
    with raises(MakolatorError, match=re.escape(r"Function 'bfunc' is not found in template")):
        mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)


def test_unknown_ignore(tmp_path):
    """Render File Inplace with missing func."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-unknown.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath, ignore_unknown=True)
    assert_refdata(test_unknown_ignore, tmp_path)


def test_child(tmp_path):
    """Render File Inplace with missing func."""
    filepath = tmp_path / "inplace-child.txt"
    copyfile(TESTDATA / "inplace-child.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace-child.txt.mako"], filepath, ignore_unknown=True)
    assert_refdata(test_child, tmp_path)


def test_mako_only(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-tpl.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_mako_only, tmp_path)


def test_mako_disabled(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-tpl.txt", filepath)
    mklt = Makolator()
    mklt.config.inplace_marker = None
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)


def test_mako_broken(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    inpfilepath = TESTDATA / "inplace-tpl-broken.txt"
    copyfile(inpfilepath, filepath)
    mklt = Makolator()
    with raises(MakolatorError, match=re.escape(" BEGIN without END.")):
        mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)

    assert filepath.read_text(encoding="utf-8") == inpfilepath.read_text(encoding="utf-8")


def test_mako_broken2(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    inpfilepath = TESTDATA / "inplace-tpl-broken2.txt"
    copyfile(inpfilepath, filepath)
    mklt = Makolator()
    with raises(CompileException, match=re.escape("Fragment")):
        mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert filepath.read_text(encoding="utf-8") == inpfilepath.read_text(encoding="utf-8")


def test_eol(tmp_path):
    """Render File Inplace Indent."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-simple.txt", filepath)
    mklt = Makolator()
    mklt.config.inplace_eol_comment = "GENERATED"
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_eol, tmp_path)


def test_eol_sv(tmp_path):
    """Render File Inplace Indent - SystemVerilog."""
    filepath = tmp_path / "inplace.sv"
    copyfile(TESTDATA / "inplace-simple.txt", filepath)
    mklt = Makolator()
    mklt.config.inplace_eol_comment = "GENERATED"
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_eol_sv, tmp_path)


def test_eol_cpp(tmp_path):
    """Render File Inplace Indent - C++."""
    filepath = tmp_path / "inplace.cpp"
    copyfile(TESTDATA / "inplace-simple.txt", filepath)
    mklt = Makolator()
    mklt.config.inplace_eol_comment = "GENERATED"
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_eol_cpp, tmp_path)


def test_eol_ini(tmp_path):
    """Render File Inplace Indent - Ini."""
    filepath = tmp_path / "inplace.ini"
    copyfile(TESTDATA / "inplace-simple.txt", filepath)
    mklt = Makolator()
    mklt.config.inplace_eol_comment = "GENERATED"
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_eol_ini, tmp_path)


def test_run(tmp_path):
    """Render File Inplace run."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-run.txt", filepath)
    mklt = Makolator()
    mklt.inplace([], filepath)
    assert_refdata(test_run, tmp_path)


def test_run_broken(tmp_path):
    """Render File Inplace run."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-run-broken.txt", filepath)
    mklt = Makolator()
    with raises(MakolatorError):
        mklt.inplace([], filepath)
    assert_paths(TESTDATA / "inplace-run-broken.txt", filepath)


def test_noend(tmp_path, mklt):
    """Render File Inplace with no end."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-noend.txt", filepath)
    with raises(MakolatorError, match=re.escape(f"'{filepath}:2' BEGIN afunc('foo')) without END.")):
        mklt.inplace([Path("inplace.txt.mako")], filepath)
    assert_paths(TESTDATA / "inplace-noend.txt", filepath)


def test_mixend(tmp_path, mklt):
    """Render File Inplace with mixed end."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-mixend.txt", filepath)
    with raises(MakolatorError, match=re.escape(f"missing END tag 'afunc' for '{filepath!s}:2'")):
        mklt.inplace([Path("inplace.txt.mako")], filepath)
    assert_paths(TESTDATA / "inplace-mixend.txt", filepath)


def test_mako_noend(tmp_path, mklt):
    """Render File Mako with no end."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-mako-noend.txt", filepath)
    with raises(MakolatorError, match=re.escape(f"'{filepath!s}:2' BEGIN without END.")):
        mklt.inplace([Path("inplace.txt.mako")], filepath)
    assert_paths(TESTDATA / "inplace-mako-noend.txt", filepath)


def test_mako_mixend(tmp_path, mklt):
    """Render File Mako with mixed end."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-mako-mixend.txt", filepath)
    with raises(MakolatorError, match=re.escape(f"missing END tag for '{filepath!s}:2'")):
        mklt.inplace([Path("inplace.txt.mako")], filepath)
    assert_paths(TESTDATA / "inplace-mako-mixend.txt", filepath)


def test_helper(tmp_path, mklt):
    """Generate File with Helper."""
    filepath = tmp_path / "helper.txt"
    copyfile(TESTDATA / "helper.txt", filepath)
    mklt.inplace([Path("helper.txt.mako")], filepath)
    assert_refdata(test_helper, tmp_path)


def test_inplace_nofill(tmp_path):
    """Render File Inplace with ffill."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-fill.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_inplace_nofill, tmp_path)


def test_inplace_fill(tmp_path):
    """Render File Inplace with ffill."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-fill.txt", filepath)
    mklt = Makolator()
    mklt.config.marker_linelength = 80
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_inplace_fill, tmp_path)


def test_inplace_fillstar(tmp_path):
    """Render File Inplace with ffill."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-fill.txt", filepath)
    mklt = Makolator()
    mklt.config.marker_fill = "*"
    mklt.config.marker_linelength = 40
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_inplace_fillstar, tmp_path)


def test_mako_only_nofill(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-tpl-fill.txt", filepath)
    mklt = Makolator()
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_mako_only_nofill, tmp_path)


def test_mako_only_fill(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-tpl-fill.txt", filepath)
    mklt = Makolator()
    mklt.config.marker_linelength = 80
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_mako_only_fill, tmp_path)


def test_mako_only_fillstar(tmp_path):
    """Render File Inplace with mako."""
    filepath = tmp_path / "inplace.txt"
    copyfile(TESTDATA / "inplace-tpl-fill.txt", filepath)
    mklt = Makolator()
    mklt.config.marker_fill = "*"
    mklt.config.marker_linelength = 40
    mklt.inplace([TESTDATA / "inplace.txt.mako"], filepath)
    assert_refdata(test_mako_only_fillstar, tmp_path)



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tracker Testing."""

from contextlib_chdir import chdir
from pytest import mark, raises
from test2ref import assert_refdata

from makolator import Config, Existing, Makolator


@mark.parametrize("existing", [Existing.KEEP, Existing.KEEP_TIMESTAMP, Existing.OVERWRITE, Existing.ERROR])
@mark.parametrize("update", ["content", "change"])
def test_tracker(tmp_path, capsys, existing, update):
    """Test Tracker."""
    mkl = Makolator(config=Config(verbose=True, existing=existing, track=True))
    with chdir(tmp_path):
        with mkl.open_outputfile("file.txt") as file:
            file.write("content")

        if existing == Existing.ERROR:
            with raises(FileExistsError):
                with mkl.open_outputfile("file.txt") as file:
                    file.write(update)
        else:
            with mkl.open_outputfile("file.txt") as file:
                file.write(update)

        with raises(RuntimeError):
            with mkl.open_outputfile("exc.txt"):
                raise RuntimeError

    tracker = mkl.tracker
    assert tracker.total == 3
    assert tracker.created == 1
    assert tracker.updated == (1 if update == "change" and existing == Existing.KEEP_TIMESTAMP else 0)
    assert tracker.identical == (1 if update == "content" and existing == Existing.KEEP_TIMESTAMP else 0)
    assert tracker.overwritten == (1 if existing == Existing.OVERWRITE else 0)
    assert tracker.identical == (1 if update == "content" and existing == Existing.KEEP_TIMESTAMP else 0)
    assert tracker.existing == (1 if existing == Existing.KEEP else 0)
    assert tracker.failed == (2 if existing == Existing.ERROR else 1)

    print(tracker.stat)

    assert_refdata(test_tracker, tmp_path, capsys=capsys, flavor=f"{existing.value}-{update}")


def test_clear(tmp_path):
    """Explicit Tracker Clear."""
    mkl = Makolator()
    mkl.config.track = True
    with chdir(tmp_path):
        with mkl.open_outputfile("file.txt") as file:
            file.write("content")

    assert mkl.tracker.stat == "1 files. 1 CREATED."
    mkl.tracker.clear()
    assert mkl.tracker.stat == "0 files."



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Makolator Testing."""

import re
import time
from pathlib import Path
from shutil import copyfile

from pytest import approx, fixture, raises
from test2ref import assert_paths, assert_refdata

from makolator import Config, Makolator, MakolatorError

FILEPATH = Path(__file__)
TESTDATA = FILEPATH.parent / "testdata"
PAUSE = 0.1


@fixture
def mklt():
    """Default :any:`Makolator` instance with proper ``template_paths``."""
    mklt = Makolator()
    mklt.config.template_paths = [TESTDATA]
    yield mklt


@fixture
def mklt_fill():
    """Default :any:`Makolator` instance with proper ``template_paths``."""
    mklt = Makolator()
    mklt.config.template_paths = [TESTDATA]
    mklt.config.marker_linelength = 94
    yield mklt


def test_abs(tmp_path, caplog, capsys):
    """Generate File With Absolute Path."""
    mklt = Makolator()
    mklt.gen([TESTDATA / "test.txt.mako"], tmp_path / "test.txt")
    assert_refdata(test_abs, tmp_path, capsys=capsys, caplog=caplog)


def test_abs_template_not_found(tmp_path):
    """Template File With Absolute Path Not Found."""
    mklt = Makolator()
    with raises(MakolatorError, match="None of the templates.*"):
        mklt.gen([TESTDATA / "test.tt.mako"], tmp_path / "test.txt")


def test_rel(tmp_path, caplog, capsys):
    """Generate File With Relative Path."""
    mklt = Makolator(config=Config(template_paths=[TESTDATA]))
    mklt.gen([Path("test.txt.mako")], tmp_path / "test.txt")
    assert_refdata(test_rel, tmp_path, capsys=capsys, caplog=caplog)


def test_rel_sub(tmp_path):
    """Generate File With Relative Path Sub."""
    mklt = Makolator(config=Config(template_paths=[TESTDATA.parent]))
    mklt.gen([Path(TESTDATA.name) / "test.txt.mako"], tmp_path / "test.txt")
    assert_refdata(test_rel_sub, tmp_path)


def test_rel_sub_not_found(tmp_path):
    """Generate File With Relative Path Sub."""
    mklt = Makolator(config=Config(template_paths=[TESTDATA.parent]))
    with raises(MakolatorError, match="None of the templates.*"):
        mklt.gen([Path(TESTDATA.name) / "test.tt.mako"], tmp_path / "test.txt")


def test_datamodel(tmp_path):
    """Use Datamodel Statement In Templates."""
    mklt = Makolator()
    mklt.datamodel.item = "myitem"
    mklt.gen([TESTDATA / "test.txt.mako"], tmp_path / "test.txt")
    assert_refdata(test_datamodel, tmp_path)


def test_datamodel_timestamp(tmp_path):
    """Keep Timestamp by Default."""
    mklt = Makolator()
    outfile = tmp_path / "test.txt"
    mklt.gen([TESTDATA / "test.txt.mako"], outfile)

    mtime = outfile.stat().st_mtime

    time.sleep(PAUSE)
    mklt.gen([TESTDATA / "test.txt.mako"], outfile)

    assert mtime == approx(outfile.stat().st_mtime)


def test_verbose(tmp_path, mklt, caplog, capsys):
    """Generate File Verbose."""
    mklt.config.verbose = True
    mklt.gen([Path("test.txt.mako")], tmp_path / "test.txt")
    mklt.gen([Path("test.txt.mako")], tmp_path / "test.txt")
    assert_refdata(test_verbose, tmp_path, caplog=caplog, capsys=capsys)


def test_stdout(tmp_path, mklt, caplog, capsys):
    """Generate File stdout."""
    mklt.gen([Path("test.txt.mako")])
    assert_refdata(test_stdout, tmp_path, caplog=caplog, capsys=capsys)


def test_context(tmp_path, mklt):
    """Generate File with Context."""
    context = {"myvar": "myvalue"}
    mklt.gen([Path("context.txt.mako")], tmp_path / "context.txt", context=context)
    assert_refdata(test_context, tmp_path)


def test_hier_base(tmp_path, mklt):
    """Generate File Hierarchy - base."""
    mklt.gen([Path("base.txt.mako")], tmp_path / "base.txt")
    assert_refdata(test_hier_base, tmp_path)


def test_hier_impl(tmp_path, mklt):
    """Generate File Hierarchy - impl."""
    mklt.gen([Path("impl.txt.mako")], tmp_path / "impl.txt")
    assert_refdata(test_hier_impl, tmp_path)


def test_run(tmp_path, mklt):
    """Use run()."""
    mklt.gen([Path("run.txt.mako")], tmp_path / "run.txt")
    assert_refdata(test_run, tmp_path)


def test_run_broken(tmp_path, mklt):
    """Use run(), which fails."""
    with raises(FileNotFoundError):
        mklt.gen([Path("run-broken.txt.mako")], tmp_path / "run.txt")


def test_static_create(tmp_path, mklt):
    """Static Code Handling."""
    filepath = tmp_path / "static.txt"
    mklt.gen([Path("static.txt.mako")], filepath)
    assert_refdata(test_static_create, tmp_path)


def test_static_create_fill(tmp_path, mklt_fill):
    """Static Code Handling."""
    filepath = tmp_path / "static.txt"
    mklt_fill.gen([Path("static.txt.mako")], filepath)
    assert_refdata(test_static_create_fill, tmp_path)


def test_static(tmp_path, mklt):
    """Static Code Handling."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static.txt", filepath)
    mklt.gen([Path("static.txt.mako")], filepath)
    assert_refdata(test_static, tmp_path)


def test_static_fill(tmp_path, mklt_fill):
    """Static Code Handling."""
    filepath = tmp_path / "static.txt"
    mklt_fill.config.marker_fill = "****"
    copyfile(TESTDATA / "static.txt", filepath)
    mklt_fill.gen([Path("static.txt.mako")], filepath)
    assert_refdata(test_static_fill, tmp_path)


def test_static_duplicate_tpl(tmp_path, mklt):
    """Static Code Handling With Duplicate In Template."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static.txt", filepath)
    with raises(MakolatorError, match=re.escape("duplicate static code 'a'")):
        mklt.gen([Path("static-duplicate.txt.mako")], filepath)
    assert_paths(TESTDATA / "static.txt", filepath)


def test_static_duplicate(tmp_path, mklt):
    """Static Code Handling With Duplicate In File."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static-duplicate.txt", filepath)
    match = re.escape(f"duplicate static code 'a' at '{filepath!s}:6'")
    with raises(MakolatorError, match=match):
        mklt.gen([Path("static.txt.mako")], filepath)
    assert_paths(TESTDATA / "static-duplicate.txt", filepath)


def test_static_noend(tmp_path, mklt):
    """Static Code Handling Without End."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static-noend.txt", filepath)
    match = re.escape(f"'{filepath!s}:2' BEGIN without END.")
    with raises(MakolatorError, match=match):
        mklt.gen([Path("static.txt.mako")], filepath)
    assert_paths(TESTDATA / "static-noend.txt", filepath)


def test_static_mixend(tmp_path, mklt):
    """Static Code Handling With Mixed End."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static-mixend.txt", filepath)
    match = re.escape(f"missing END tag 'a' for '{filepath!s}:2'")
    with raises(MakolatorError, match=match):
        mklt.gen([Path("static.txt.mako")], filepath)
    assert_paths(TESTDATA / "static-mixend.txt", filepath)


def test_static_unknown(tmp_path, mklt):
    """Static Code Handling With Unknown."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static-unknown.txt", filepath)
    match = re.escape(f"'{filepath!s}': unknown static code 'c'")
    with raises(MakolatorError, match=match):
        mklt.gen([Path("static.txt.mako")], filepath)
    assert_paths(TESTDATA / "static-unknown.txt", filepath)


def test_static_corner_create(tmp_path, mklt, caplog):
    """Generate File with Corner Cases."""
    filepath = tmp_path / "static.txt"
    mklt.gen([Path("static-corner.txt.mako")], filepath)
    assert_refdata(test_static_corner_create, tmp_path, caplog=caplog)


def test_static_corner(tmp_path, mklt, caplog):
    """Generate File with Corner Cases."""
    filepath = tmp_path / "static.txt"
    copyfile(TESTDATA / "static-corner.txt", filepath)
    mklt.gen([Path("static-corner.txt.mako")], filepath)
    assert_refdata(test_static_corner, tmp_path, caplog=caplog)


def test_helper(tmp_path, mklt):
    """Generate File with Helper."""
    filepath = tmp_path / "helper.txt"
    mklt.gen([Path("helper.txt.mako")], filepath)
    assert_refdata(test_helper, tmp_path)


def test_undefined(tmp_path, mklt):
    """Generate File with Undefined."""
    filepath = tmp_path / "undefined.txt"
    with raises(NameError):
        mklt.gen([Path("undefined.txt.mako")], filepath)


def test_gen_recursive(tmp_path):
    """Recursive Rendering."""
    mklt = Makolator()
    mklt.datamodel.name = "some-name"
    mklt.gen([TESTDATA / "gen-recursive"], tmp_path)
    assert_refdata(test_gen_recursive, tmp_path)



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Escape reserved characters."""

import os
import re

__TEX_CONV = {
    "&": r"\&",
    "%": r"\%",
    "$": r"\$",
    "#": r"\#",
    "_": r"\_",
    "{": r"\{",
    "}": r"\}",
    "~": r"\textasciitilde{}",
    "^": r"\^{}",
    "\\": r"\textbackslash{}",
    "<": r"\textless{}",
    ">": r"\textgreater{}",
    "®": r"\textsuperscript{\textregistered}",
    "©": r"\textcopyright{}",
    "™": r"\textsuperscript{\texttrademark}",
}
__TEX_REGEX = re.compile("|".join(re.escape(key) for key in sorted(__TEX_CONV.keys(), key=lambda item: -len(item))))


def tex(text: str | None):
    r"""
    Escape (La)Tex.

    Args:
        text: tex
    Returns:
        escaped string.

        >>> tex("Foo & Bar")
        'Foo \\& Bar'
        >>> tex(None)
    """
    if text is None:
        return None
    result = __TEX_REGEX.sub(lambda match: __TEX_CONV[match.group()], str(text))
    if os.linesep != "\n":
        return result.replace("\n", os.linesep)
    return result



#
# MIT License
#
# Copyright (c) 2023-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Makolator Testing."""

from pathlib import Path

from contextlib_chdir import chdir

from makolator import Config, Datamodel, Existing, Makolator, Tracker

FILEPATH = Path(__file__)
TESTDATA = FILEPATH.parent / "testdata"


def test_makolator():
    """Basic Testing On Makolator."""
    mkl = Makolator()
    assert mkl.config == Config()
    assert mkl.datamodel == Datamodel()
    assert mkl.tracker == Tracker()


def test_outputfile(tmp_path, capsys):
    """Test Outputfile."""
    mkl = Makolator()
    with chdir(tmp_path):
        with mkl.open_outputfile("file.txt") as file:
            file.write("content")
        assert file.state.name == "CREATED"
        with mkl.open_outputfile("file.txt") as file:
            file.write("change")
        assert file.state.name == "UPDATED"

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""
    assert mkl.tracker.total == 0


def test_outputfile_keep(tmp_path, capsys):
    """Test Outputfile With Keep."""
    mkl = Makolator()
    mkl.config.existing = Existing.KEEP
    with chdir(tmp_path):
        with mkl.open_outputfile("file.txt") as file:
            file.write("content")
        assert file.state.name == "CREATED"
        with mkl.open_outputfile("file.txt") as file:
            file.write("change")
        assert file.state.name == "EXISTING"

    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""
    assert mkl.tracker.total == 0


def test_create_dir(tmp_path):
    """Create Output Directory Structure If It Does Not Exist."""
    mklt = Makolator()
    outfile = tmp_path / "sub" / "test.txt"
    mklt.gen([TESTDATA / "test.txt.mako"], outfile)


def test_cachepath_default(tmp_path):
    """Default Cache Path."""
    mklt = Makolator()
    cachepath = mklt.cache_path
    assert cachepath.exists()
    assert len(tuple(cachepath.glob("*"))) == 0
    mklt.gen([TESTDATA / "test.txt.mako"], tmp_path / "test.txt")


def test_cachepath_explicit(tmp_path):
    """Explicit Cache Path."""
    mklt = Makolator(config=Config(cache_path=tmp_path / "cache"))
    cachepath = mklt.cache_path
    assert cachepath.exists()
    assert len(tuple(cachepath.glob("*"))) == 0
    mklt.gen([TESTDATA / "test.txt.mako"], tmp_path / "test.txt")
